@@ -1,5 +1,5 @@
 # -*- coding: utf-8 -*-
-# Copyright (c) 2013, imageio contributers
+# Copyright (c) 2014, imageio contributors
 # imageio is distributed under the terms of the (new) BSD License.
 
 """ Plugin that uses ffmpeg to read and write series of images to
@@ -10,32 +10,50 @@ by Zulko
 
 """
 
+from __future__ import absolute_import, print_function, division
+
 import sys
 import os
+import stat
 import re
 import time
 import threading
+import struct
 import subprocess as sp
 
 import numpy as np
 
-import imageio
 from imageio import formats
-from imageio.base import Format
-
-# todo: select an exe!
-FFMPEG_EXE = os.path.join(os.path.dirname(imageio.__file__), 'lib', 'ffmpeg')
-if sys.platform.startswith('win'):
-    FFMPEG_EXE += '.exe'
+from imageio.core import Format, get_remote_file, string_types, read_n_bytes
 
 
+def get_exe():
+    """ Get ffmpeg exe
+    """
+    NBYTES = struct.calcsize('P') * 8
+    if sys.platform.startswith('linux'):
+        fname = 'ffmpeg.linux%i' % NBYTES
+    elif sys.platform.startswith('win'):
+        fname = 'ffmpeg.win32.exe'
+    elif sys.platform.startswith('darwin'):
+        fname = 'ffmpeg.osx.snowleopardandabove'
+    else:  # pragma: no cover
+        fname = 'ffmpeg'  # hope for the best
+    #
+    FFMPEG_EXE = 'ffmpeg'
+    if fname:
+        FFMPEG_EXE = get_remote_file('ffmpeg/' + fname)
+        os.chmod(FFMPEG_EXE, os.stat(FFMPEG_EXE).st_mode | stat.S_IEXEC)  # exe
+    return FFMPEG_EXE
+
+# Get camera format
 if sys.platform.startswith('win'):
     CAM_FORMAT = 'dshow'  # dshow or vfwcap
 elif sys.platform.startswith('linux'):
-    CAM_FORMAT ='video4linux2'
+    CAM_FORMAT = 'video4linux2'
 elif sys.platform.startswith('darwin'):
-    CAM_FORMAT ='??'
-else:
+    CAM_FORMAT = '??'
+else:  # pragma: no cover
     CAM_FORMAT = 'unknown-cam-format'
 
 
@@ -48,30 +66,40 @@ class FfmpegFormat(Format):
     where the "0" can be replaced with any index of cameras known to
     the system.
     
-    Keyword arguments for reading
-    -----------------------------
+    Note that for reading regular video files, the avbin plugin is more
+    efficient.
+    
+    Parameters for reading
+    ----------------------
     loop : bool
         If True, the video will rewind as soon as a frame is requested
-        beyond the last frame. Otherwise, IndexError is raised.
-    size : str
-        The frame size (i.e. resolution) to read the images, e.g. "640x480". 
-        For camera streams, this allows setting the capture resolution. 
-        For normal video data, ffmpeg will rescale the data.
+        beyond the last frame. Otherwise, IndexError is raised. Default False.
+    size : str | tuple
+        The frame size (i.e. resolution) to read the images, e.g. 
+        (100, 100) or "640x480". For camera streams, this allows setting
+        the capture resolution. For normal video data, ffmpeg will
+        rescale the data.
     pixelformat : str
         The pixel format for the camera to use (e.g. "yuyv422" or
         "gray"). The camera needs to support the format in order for
         this to take effect. Note that the images produced by this
         reader are always rgb8.
+    print_info : bool
+        Print information about the video file as reported by ffmpeg.
     
+    Parameters for saving
+    ---------------------
+    fps : scalar
+        The number of frames per second. Default 10.
+    codec : str
+        the video codec to use. Default 'libx264', which represents the
+        widely available mpeg4.
+    bitrate : int
+        A measure for quality. Default 400000
     """
     
     def _can_read(self, request):
-        # The request object has:
-        # request.filename: the filename
-        # request.firstbytes: the first 256 bytes of the file.
-        # request.expect: what kind of data the user expects
-        # request.kwargs: the keyword arguments specified by the user
-        if (request.expect is not None) and request.expect != imageio.EXPECT_MIM:
+        if request.mode[1] not in 'I?':
             return False
         
         # Read from video stream?
@@ -81,13 +109,17 @@ class FfmpegFormat(Format):
             return True
         
         # Read from file that we know?
-        ext = os.path.splitext(request.filename)[1].lower()
-        if ext in '.avi .mpg .mpeg .mp4':
-            return True
+        for ext in self.extensions:
+            if request.filename.endswith('.' + ext):
+                return True
     
     def _can_save(self, request):
-        return self._can_read(request)
+        if request.mode[1] in (self.modes + '?'):
+            for ext in self.extensions:
+                if request.filename.endswith('.' + ext):
+                    return True
     
+    # --
     
     class Reader(Format.Reader):
         
@@ -97,9 +129,10 @@ class FfmpegFormat(Format):
             
             elif sys.platform.startswith('win'):
                 # Ask ffmpeg for list of dshow device names
-                cmd = [FFMPEG_EXE, '-list_devices', 'true',
-                                '-f', CAM_FORMAT, '-i', 'dummy']
-                proc = sp.Popen(cmd, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE)
+                cmd = [self._exe, '-list_devices', 'true',
+                       '-f', CAM_FORMAT, '-i', 'dummy']
+                proc = sp.Popen(cmd, stdin=sp.PIPE, stdout=sp.PIPE, 
+                                stderr=sp.PIPE)
                 proc.stdout.readline()
                 proc.terminate()
                 infos = proc.stderr.read().decode('utf-8')
@@ -122,14 +155,28 @@ class FfmpegFormat(Format):
                 except IndexError:
                     raise IndexError('No ffdshow camera at index %i.' % index)
                 return 'video=%s' % name
-            else:
+            else:  # pragma: no cover
                 return '??'
         
-        def _open(self, loop=False, size=None, pixelformat=None):
+        def _open(self, loop=False, size=None, pixelformat=None, 
+                  print_info=False):
+            # Get exe
+            self._exe = get_exe()
             # Process input args
-            self._arg_loop = loop
-            self._arg_size = size
-            self._arg_pxelformat = pixelformat
+            self._arg_loop = bool(loop)
+            if size is None:
+                self._arg_size = None
+            elif isinstance(size, tuple):
+                self._arg_size = "%ix%i" % size
+            elif isinstance(size, string_types) and 'x' in size:
+                self._arg_size = size
+            else:
+                raise ValueError('FFMPEG size must be tuple of "NxM"')
+            if pixelformat is None:
+                pass
+            elif not isinstance(pixelformat, string_types):
+                raise ValueError('FFMPEG pixelformat must be str')
+            self._arg_pixelformat = pixelformat
             # Write "_video"_arg
             self.request._video = None
             if self.request.filename in ['<video%i>' % i for i in range(10)]:
@@ -142,11 +189,12 @@ class FfmpegFormat(Format):
                 self._filename = self.request.get_local_filename()
             # Determine pixel format and depth
             self._pix_fmt = 'rgb24'
-            self._depth = 4 if self._pix_fmt=="rgba" else 3
+            self._depth = 4 if self._pix_fmt == "rgba" else 3
             # Initialize parameters
             self._proc = None
             self._pos = -1
-            self._meta = {'nframes': float('inf'), 'nframes': float('inf')}
+            self._meta = {'plugin': 'ffmpeg', 
+                          'nframes': float('inf'), 'nframes': float('inf')}
             self._lastread = None
             # Start ffmpeg subprocess and get meta information
             self._initialize()
@@ -157,10 +205,11 @@ class FfmpegFormat(Format):
             if self.request._video:
                 w, h = self._meta['size']
                 framesize = self._depth * w * h
-                self._frame_catcher = FrameCatcher(self._proc.stdout, framesize)
+                self._frame_catcher = FrameCatcher(self._proc.stdout, 
+                                                   framesize)
         
         def _close(self):
-            self._terminate()
+            self._terminate(0.05)  # Short timeout
             self._proc = None
         
         def _get_length(self):
@@ -179,6 +228,10 @@ class FfmpegFormat(Format):
             
             if index == self._pos:
                 return self._lastread, {}
+            elif index < 0:
+                raise IndexError('Frame index must be > 0') 
+            elif index >= self._meta['nframes']:
+                raise IndexError('Reached end of video')
             else:
                 if (index < self._pos) or (index > self._pos+100):
                     self._reinitialize(index)
@@ -191,19 +244,13 @@ class FfmpegFormat(Format):
         def _get_meta_data(self, index):
             return self._meta
         
-        def _get_next_data(self):
-            result = self._read_frame()
-            self._pos += 1
-            return result, {}
-        
-        
         def _initialize(self):
             """ Opens the file, creates the pipe. """
             # Create input args
             if self.request._video:
                 iargs = ['-f', CAM_FORMAT]
-                if self._arg_pxelformat:
-                    iargs.extend(['-pix_fmt', self._arg_pxelformat])
+                if self._arg_pixelformat:
+                    iargs.extend(['-pix_fmt', self._arg_pixelformat])
                 if self._arg_size:
                     iargs.extend(['-s', self._arg_size])
             else:
@@ -212,21 +259,19 @@ class FfmpegFormat(Format):
             oargs = ['-f', 'image2pipe',
                      '-pix_fmt', self._pix_fmt,
                      '-vcodec', 'rawvideo']
-            if 'size' in self.request.kwargs:
-                oargs.extend(['-s', self.request.kwargs['size']])
+            oargs.extend(['-s', self._arg_size] if self._arg_size else [])
             # Create process
-            cmd = [FFMPEG_EXE] + iargs + ['-i', self._filename] + oargs + ['-']
+            cmd = [self._exe] + iargs + ['-i', self._filename] + oargs + ['-']
             self._proc = sp.Popen(cmd, stdin=sp.PIPE,
                                   stdout=sp.PIPE, stderr=sp.PIPE)
             # Create thread that keeps reading from stderr
             self._stderr_catcher = StreamCatcher(self._proc.stderr)
         
-        
         def _reinitialize(self, index=0):
             """ Restarts the reading, starts at an arbitrary location
             (!! SLOW !!) """
             if self.request._video:
-                raise RuntimeError('No random access when streaming from camera.')
+                raise RuntimeError('No random access when streaming from cam.')
             self._close()
             if index == 0:
                 self._initialize()
@@ -234,40 +279,46 @@ class FfmpegFormat(Format):
                 starttime = index / self._meta['fps']
                 offset = min(1, starttime)
                 # Create input args -> start time
-                iargs = ['-ss', "%.03f"%(starttime-offset)]
+                iargs = ['-ss', "%.03f" % (starttime-offset)]
                 # Output args, for writing to pipe
                 oargs = ['-f', 'image2pipe',
-                        '-pix_fmt', self._pix_fmt,
-                        '-vcodec', 'rawvideo']
-                if 'size' in self.request.kwargs:
-                    oargs.extend(['-s', self.request.kwargs['size']])
+                         '-pix_fmt', self._pix_fmt,
+                         '-vcodec', 'rawvideo']
+                oargs.extend(['-s', self._arg_size] if self._arg_size else [])
                 # Create process
-                cmd = [FFMPEG_EXE] + iargs + ['-i', self._filename] + oargs + ['-']
+                cmd = [self._exe]
+                cmd += iargs + ['-i', self._filename] + oargs + ['-']
                 self._proc = sp.Popen(cmd, stdin=sp.PIPE,
                                       stdout=sp.PIPE, stderr=sp.PIPE)
                 # Create thread that keeps reading from stderr
                 self._stderr_catcher = StreamCatcher(self._proc.stderr)
         
-        
-        def _terminate(self):
+        def _terminate(self, timeout=1.0):
             """ Terminate the sub process.
             """
             # Check
-            if self._proc is None:
+            if self._proc is None:  # pragma: no cover
                 return  # no process
             if self._proc.poll() is not None:
                 return  # process already dead
-            # Close, terminate
-            for std in (self._proc.stdin, self._proc.stdout, self._proc.stderr):
-                std.close()
+            # Terminate process
             self._proc.terminate()
             # Wait for it to close (but do not get stuck)
-            etime = time.time() + 1.0
+            etime = time.time() + timeout
             while time.time() < etime:
                 time.sleep(0.01)
                 if self._proc.poll() is not None:
                     break
         
+#         def _close_streams(self):
+#             for std in (self._proc.stdin, 
+#                         self._proc.stdout, 
+#                         self._proc.stderr):
+#                 try:
+#                     std.close()
+#                 except Exception:  # pragma: no cover
+#                     pass
+        
         def _load_infos(self):
             """ reads the FFMPEG info on the file and sets size fps
             duration and nframes. """
@@ -281,11 +332,15 @@ class FfmpegFormat(Format):
             infos = self._stderr_catcher.header
             if not infos:
                 self._terminate()
-                err2 = self._stderr_catcher.get_text(0.2)
-                fmt = 'Could not load meta information\n=== stderr ===\n%s'
-                raise RuntimeError(fmt % err2)
+                if self.request._video:
+                    raise IndexError('No video4linux camera at %s.' % 
+                                     self.request._video)
+                else:
+                    err2 = self._stderr_catcher.get_text(0.2)
+                    fmt = 'Could not load meta information\n=== stderr ===\n%s'
+                    raise IOError(fmt % err2)
             
-            if self.request.kwargs.get('print_infos', False):
+            if self.request.kwargs.get('print_info', False):
                 # print the whole info text returned by FFMPEG
                 print(infos)
                 print('-'*80)
@@ -293,7 +348,7 @@ class FfmpegFormat(Format):
             if "No such file or directory" in lines[-1]:
                 if self.request._video:
                     raise IOError("Could not open steam %s." % self._filename)
-                else:
+                else:  # pragma: no cover - this is checked by Request
                     raise IOError("%s not found! Wrong path?" % self._filename)
             
             # Get version
@@ -305,17 +360,20 @@ class FfmpegFormat(Format):
             line = videolines[0]
             
             # get the frame rate
-            match = re.search("( [0-9]*.| )[0-9]* (tbr|fps)",line)
-            self._meta['fps'] = fps = float(line[match.start():match.end()].split(' ')[1])
+            match = re.search("( [0-9]*.| )[0-9]* (tbr|fps)", line)
+            fps = float(line[match.start():match.end()].split(' ')[1])
+            self._meta['fps'] = fps
             
             # get the size of the original stream, of the form 460x320 (w x h)
             match = re.search(" [0-9]*x[0-9]*(,| )", line)
-            self._meta['source_size'] = tuple(map(int,line[match.start():match.end()-1].split('x')))
+            parts = line[match.start():match.end()-1].split('x')
+            self._meta['source_size'] = tuple(map(int, parts))
             
             # get the size of what we receive, of the form 460x320 (w x h)
             line = videolines[-1]  # Pipe output
             match = re.search(" [0-9]*x[0-9]*(,| )", line)
-            self._meta['size'] = tuple(map(int,line[match.start():match.end()-1].split('x')))
+            parts = line[match.start():match.end()-1].split('x')
+            self._meta['size'] = tuple(map(int, parts))
             
             # Check the two sizes
             if self._meta['source_size'] != self._meta['size']:
@@ -325,9 +383,10 @@ class FfmpegFormat(Format):
             
             # get duration (in seconds)
             line = [l for l in lines if 'Duration: ' in l][0]
-            match = re.search(" [0-9][0-9]:[0-9][0-9]:[0-9][0-9].[0-9][0-9]",line)
+            match = re.search(" [0-9][0-9]:[0-9][0-9]:[0-9][0-9].[0-9][0-9]",
+                              line)
             if match is not None:
-                hms = map(float,line[match.start()+1:match.end()].split(':'))
+                hms = map(float, line[match.start()+1:match.end()].split(':'))
                 self._meta['duration'] = duration = cvsecs(*hms)
                 self._meta['nframes'] = int(duration*fps)
         
@@ -335,26 +394,17 @@ class FfmpegFormat(Format):
             # Init and check
             w, h = self._meta['size']
             framesize = self._depth * w * h
-            if self._proc is None:
-                raise RuntimeError('No active ffmpeg process, maybe the reader is closed?')
+            assert self._proc is not None
             
             try:
                 # Read framesize bytes
-                if self._frame_catcher:
+                if self._frame_catcher:  # pragma: no cover - camera thing
                     s = self._frame_catcher.get_frame()
                 else:
-                    s = self._proc.stdout.read(framesize)
-                    while len(s) < framesize:
-                        need = framesize - len(s)
-                        part = self._proc.stdout.read(need)
-                        if not part:
-                            break
-                        s += part
+                    s = read_n_bytes(self._proc.stdout, framesize)
                 # Check
                 assert len(s) == framesize
             except Exception as err:
-                if not s:
-                    raise IndexError('Stream ended')
                 self._terminate()
                 err1 = str(err)
                 err2 = self._stderr_catcher.get_text(0.4)
@@ -372,20 +422,23 @@ class FfmpegFormat(Format):
         def _read_frame(self):
             # Read and convert to numpy array
             w, h = self._meta['size']
-            t0 = time.time()
+            #t0 = time.time()
             s = self._read_frame_data()
-            result = np.fromstring(s, dtype='uint8').reshape((h,w,self._depth))
-            t1 = time.time()
+            result = np.fromstring(s, dtype='uint8')
+            result = result.reshape((h, w, self._depth))
+            #t1 = time.time()
             #print('etime', t1-t0)
             
             # Store and return
             self._lastread = result
             return result
     
+    # --
     
     class Writer(Format.Writer):
         
-        def _open(self):        
+        def _open(self, fps=10, codec='libx264', bitrate=400000):
+            self._exe = get_exe()
             # Get local filename
             self._filename = self.request.get_local_filename()
             # Determine pixel format and depth
@@ -394,7 +447,6 @@ class FfmpegFormat(Format):
             self._proc = None
             self._size = None
         
-        
         def _close(self):
             # Close subprocess
             if self._proc is not None:
@@ -410,33 +462,34 @@ class FfmpegFormat(Format):
             depth = 1 if im.ndim == 2 else im.shape[2]
             
             # Ensure that image is in uint8
-            im = im.astype(np.uint8, copy=False)
+            if im.dtype != np.uint8:
+                im = im.astype(np.uint8)  # pypy: no support copy=False
             
             # Set size and initialize if not initialized yet
             if self._size is None:
-                self._size = size
-                self._depth = depth
-                map = {1:'gray', 2:'gray8a', 3:'rgb24', 4:'rgba'}
+                map = {1: 'gray', 2: 'gray8a', 3: 'rgb24', 4: 'rgba'}
                 self._pix_fmt = map.get(depth, None)
                 if self._pix_fmt is None:
-                    raise ValueError('Image data must have 1, 2, 3 or 4 channels.')
+                    raise ValueError('Image must have 1, 2, 3 or 4 channels')
+                self._size = size
+                self._depth = depth
                 self._initialize()
             
             # Check size of image
             if size != self._size:
-                raise ValueError('All images in a movie should have the same size.')
+                raise ValueError('All images in a movie should have same size')
             if depth != self._depth:
-                raise ValueError('All images in a movie should have the same number of channels.')
+                raise ValueError('All images in a movie should have same '
+                                 'number of channels')
             
-            # Check status
-            if self._proc is None:
-                raise RuntimeError('No active ffmpeg process, maybe writer is closed?')
+            assert self._proc is not None  # Check status
             
             # Write
             self._proc.stdin.write(im.tostring())
         
         def set_meta_data(self, meta):
-            raise RuntimeError('The ffmpeg format does not support setting meta data.')
+            raise RuntimeError('The ffmpeg format does not support setting '
+                               'meta data.')
         
         def _initialize(self):
             """ Creates the pipe to ffmpeg. Open the file to write to. """
@@ -446,19 +499,19 @@ class FfmpegFormat(Format):
             # do not specify a bitrate, we easily get crap results.
             sizestr = "%dx%d" % (self._size[1], self._size[0])
             fps = self.request.kwargs.get('fps', 10)
-            codec = self.request.kwargs.get('codec', 'libx264') # libx264 mpeg4 msmpeg4v2
+            codec = self.request.kwargs.get('codec', 'libx264')
             bitrate = self.request.kwargs.get('bitrate', 400000)
             
             # Get command
-            cmd = [FFMPEG_EXE, '-y',
-                "-f", 'rawvideo',
-                "-vcodec", "rawvideo",
-                '-s', sizestr,
-                '-pix_fmt', self._pix_fmt,
-                '-r', "%.02f" % fps,
-                '-i', '-', '-an',
-                '-vcodec', codec] 
-            cmd += ['-b', str(bitrate)] if (bitrate!=None) else [] 
+            cmd = [self._exe, '-y',
+                   "-f", 'rawvideo',
+                   "-vcodec", "rawvideo",
+                   '-s', sizestr,
+                   '-pix_fmt', self._pix_fmt,
+                   '-r', "%.02f" % fps,
+                   '-i', '-', '-an',
+                   '-vcodec', codec] 
+            cmd += ['-b', str(bitrate)] if (bitrate is not None) else [] 
             cmd += ['-r', "%d" % fps, self._filename]
             
             # Launch process
@@ -467,17 +520,24 @@ class FfmpegFormat(Format):
 
 
 def cvsecs(*args):
-    """ converts a time to second. Either cvsecs(min,secs) or
-    cvsecs(hours,mins,secs).
+    """ converts a time to second. Either cvsecs(min, secs) or
+    cvsecs(hours, mins, secs).
     """
     if len(args) == 1:
         return args[0]
     elif len(args) == 2:
-        return 60*args[0]+args[1]
-    elif len(args) ==3:
-        return 3600*args[0]+60*args[1]+args[2]
+        return 60*args[0] + args[1]
+    elif len(args) == 3:
+        return 3600*args[0] + 60*args[1] + args[2]
 
 
+def limit_lines(lines, N=32):
+    """ When number of lines > 2*N, reduce to N.
+    """
+    if len(lines) > 2*N:
+        lines = [b'... showing only last few lines ...'] + lines[-N:]
+    return lines
+
 
 class FrameCatcher(threading.Thread):
     """ Thread to keep reading the frame data from stdout. This is
@@ -491,13 +551,14 @@ class FrameCatcher(threading.Thread):
         self._file = file
         self._framesize = framesize
         self._frame = None
+        self._bytes_read = 0
         #self._lock = threading.RLock()
         threading.Thread.__init__(self)
         self.setDaemon(True)  # do not let this thread hold up Python shutdown
         self.start()
     
     def get_frame(self):
-        while self._frame is None:
+        while self._frame is None:  # pragma: no cover - an init thing
             time.sleep(0.001)
         return self._frame
     
@@ -519,6 +580,7 @@ class FrameCatcher(threading.Thread):
                 if not part:
                     break
                 s += part
+                self._bytes_read += len(part)
             # Stop?
             if not s:
                 return
@@ -526,7 +588,6 @@ class FrameCatcher(threading.Thread):
             self._frame = s
 
 
-
 class StreamCatcher(threading.Thread):
     """ Thread to keep reading from stderr so that the buffer does not
     fill up and stalls the ffmpeg process. On stderr a message is send
@@ -561,7 +622,7 @@ class StreamCatcher(threading.Thread):
         # Wait?
         if timeout > 0:
             etime = time.time() + timeout
-            while self.isAlive() and time.time() < etime:
+            while self.isAlive() and time.time() < etime:  # pragma: no cover
                 time.sleep(0.025)
         # Return str
         lines = b'\n'.join(self._lines)
@@ -573,7 +634,7 @@ class StreamCatcher(threading.Thread):
             # Read one line. Detect when closed, and exit
             try:
                 line = self._file.read(20)
-            except ValueError:
+            except ValueError:  # pragma: no cover
                 break
             if not line:
                 break
@@ -589,14 +650,10 @@ class StreamCatcher(threading.Thread):
                     header = b'\n'.join(self._lines)
                     self._header += header.decode('utf-8', 'ignore')
                     self._lines = []
-            # Check lines
-            N = 32
-            if len(self._lines) > 2*N:
-                self._lines = [b'... showing only last few lines ...'] + self._lines[N:]
-            
+            self._lines = limit_lines(self._lines)
 
 
 # Register. You register an *instance* of a Format class.
-format = FfmpegFormat('ffmpeg', 'Many video formats and cameras via ffmpeg.', 
-                      '.avi .mpg .mpeg .mp4')
+format = FfmpegFormat('ffmpeg', 'Many video formats and cameras (via ffmpeg)', 
+                      '.mov .avi .mpg .mpeg .mp4', 'I')
 formats.add_format(format)
